// ****************************
// * Hacksmith Servers Plugin *
// ****************************
// version 3.0
// by Pierre-Marie Baty <pm@bots-united.com>


#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>

enginefuncs_t g_engfuncs;
globalvars_t  *gpGlobals;
DLL_FUNCTIONS gFunctionTable;
cvar_t acknowledgment_string = {"acknowledgment_string", "acknowledged", FCVAR_PROTECTED};
cvar_t player_force_team = {"player_force_team", "2", FCVAR_EXTDLL};
cvar_t send_map_message = {"send_map_message", "1", FCVAR_EXTDLL};
bool acknowledged_players[33];
float acknowledgment_time[33];
bool ignore_message = FALSE;
bool round_end = FALSE;
char welcome_text[512] = "Welcome to this server.\n"
                         "\n"
                         "You can display here your own general conditions\n"
                         "that you can write in the welcome.msg file\n"
                         "in the cstrike directory.\n"
                         "\n"
                         "Bring down the console and type in \"acknowledged\"\n"
                         "to start the game.\n";
char map_text[512] = "You can display here your own\n"
                     "map tactics and advices for both teams\n"
                     "that you can write in the MAPNAME.msg file\n"
                     "in the cstrike\\maps directory.\n";

void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals);
void ClientDisconnect (edict_t *pEntity);
void ClientPutInServer (edict_t *pEntity);
void ClientCommand (edict_t *pEntity);
void StartFrame (void);
void engClientCommand (edict_t *pEdict, char *szFmt, ...);
void MessageBegin (int msg_dest, int msg_type, const float *pOrigin, edict_t *ed);
void MessageEnd (void);
void WriteByte (int iValue);
void WriteChar (int iValue);
void WriteShort (int iValue);
void WriteLong (int iValue);
void WriteAngle (float flValue);
void WriteCoord (float flValue);
void WriteString (const char *sz);
void WriteEntity (int iValue);
void SendWelcomeMessage (edict_t *pClient);
void SendMapMessage (void);
void GetGameDir (char *game_dir);


// START of Metamod stuff

enginefuncs_t meta_engfuncs;
gamedll_funcs_t *gpGamedllFuncs;
mutil_funcs_t *gpMetaUtilFuncs;
meta_globals_t *gpMetaGlobals;

META_FUNCTIONS gMetaFunctionTable =
{
   NULL, // pfnGetEntityAPI()
   NULL, // pfnGetEntityAPI_Post()
   GetEntityAPI2, // pfnGetEntityAPI2()
   NULL, // pfnGetEntityAPI2_Post()
   NULL, // pfnGetNewDLLFunctions()
   NULL, // pfnGetNewDLLFunctions_Post()
   GetEngineFunctions, // pfnGetEngineFunctions()
   NULL, // pfnGetEngineFunctions_Post()
};

plugin_info_t Plugin_info = {
   META_INTERFACE_VERSION, // interface version
   "Hacksmith", // plugin name
   "3.0", // plugin version
   "13/02/2004", // date of creation
   "Pierre-Marie Baty <pm@bots-united.com>", // plugin author
   "http://racc.bots-united.com", // plugin URL
   "HACKSMITH", // plugin logtag
   PT_ANYTIME, // when loadable
   PT_ANYTIME, // when unloadable
};


C_DLLEXPORT int Meta_Query (char *ifvers, plugin_info_t **pPlugInfo, mutil_funcs_t *pMetaUtilFuncs)
{
   // this function is the first function ever called by metamod in the plugin DLL. Its purpose
   // is for metamod to retrieve basic information about the plugin, such as its meta-interface
   // version, for ensuring compatibility with the current version of the running metamod.

   // keep track of the pointers to metamod function tables metamod gives us
   gpMetaUtilFuncs = pMetaUtilFuncs;
   *pPlugInfo = &Plugin_info;

   // check for interface version compatibility
   if (strcmp (ifvers, Plugin_info.ifvers) != 0)
   {
      int mmajor = 0, mminor = 0, pmajor = 0, pminor = 0;

      LOG_CONSOLE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);
      LOG_MESSAGE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);

      // if plugin has later interface version, it's incompatible (update metamod)
      sscanf (ifvers, "%d:%d", &mmajor, &mminor);
      sscanf (META_INTERFACE_VERSION, "%d:%d", &pmajor, &pminor);

      if ((pmajor > mmajor) || ((pmajor == mmajor) && (pminor > mminor)))
      {
         LOG_CONSOLE (PLID, "metamod version is too old for this plugin; update metamod");
         LOG_ERROR (PLID, "metamod version is too old for this plugin; update metamod");
         return (FALSE);
      }

      // if plugin has older major interface version, it's incompatible (update plugin)
      else if (pmajor < mmajor)
      {
         LOG_CONSOLE (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         LOG_ERROR (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         return (FALSE);
      }
   }

   return (TRUE); // tell metamod this plugin looks safe
}


C_DLLEXPORT int Meta_Attach (PLUG_LOADTIME now, META_FUNCTIONS *pFunctionTable, meta_globals_t *pMGlobals, gamedll_funcs_t *pGamedllFuncs)
{
   // this function is called when metamod attempts to load the plugin. Since it's the place
   // where we can tell if the plugin will be allowed to run or not, we wait until here to make
   // our initialization stuff, like registering CVARs and dedicated server commands.

   char dirname[256];

   // are we allowed to load this plugin now ?
   if (now > Plugin_info.loadable)
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

   // keep track of the pointers to engine function tables metamod gives us
   gpMetaGlobals = pMGlobals;
   memcpy (pFunctionTable, &gMetaFunctionTable, sizeof (META_FUNCTIONS));
   gpGamedllFuncs = pGamedllFuncs;

   // get the MOD directory name (full path)
   GetGameDir (dirname);

   // is it NOT the Counter-Strike directory ?
   if (strcmp (dirname, "cstrike") != 0)
   {
      // then don't allow this plugin to run
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (bad dirname: '%s')", Plugin_info.name, dirname);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (bad dirname: '%s')", Plugin_info.name, dirname);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

   // print a message to notify about plugin attaching
   LOG_CONSOLE (PLID, "%s: plugin attaching", Plugin_info.name);
   LOG_MESSAGE (PLID, "%s: plugin attaching", Plugin_info.name);

   // ask the engine to register the CVARs this plugin uses
   CVAR_REGISTER (&acknowledgment_string);
   CVAR_REGISTER (&player_force_team);
   CVAR_REGISTER (&send_map_message);

   return (TRUE); // returning TRUE enables metamod to attach this plugin
}


C_DLLEXPORT int Meta_Detach (PLUG_LOADTIME now, PL_UNLOAD_REASON reason)
{
   // this function is called when metamod unloads the plugin. A basic check is made in order
   // to prevent unloading the plugin if its processing should not be interrupted.

   // is metamod allowed to unload the plugin ?
   if ((now > Plugin_info.unloadable) && (reason != PNL_CMD_FORCED))
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from unloading this plugin
   }

   return (TRUE); // returning TRUE enables metamod to unload this plugin
}


C_DLLEXPORT int GetEntityAPI2 (DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion)
{
   gFunctionTable.pfnClientDisconnect = ClientDisconnect;
   gFunctionTable.pfnClientPutInServer = ClientPutInServer;
   gFunctionTable.pfnClientCommand = ClientCommand;
   gFunctionTable.pfnStartFrame = StartFrame;

   memcpy (pFunctionTable, &gFunctionTable, sizeof (DLL_FUNCTIONS));
   return (TRUE);
}


C_DLLEXPORT int GetEngineFunctions (enginefuncs_t *pengfuncsFromEngine, int *interfaceVersion)
{
   meta_engfuncs.pfnClientCommand = engClientCommand;
   meta_engfuncs.pfnMessageBegin = MessageBegin;
   meta_engfuncs.pfnMessageEnd = MessageEnd;
   meta_engfuncs.pfnWriteByte = WriteByte;
   meta_engfuncs.pfnWriteChar = WriteChar;
   meta_engfuncs.pfnWriteShort = WriteShort;
   meta_engfuncs.pfnWriteLong = WriteLong;
   meta_engfuncs.pfnWriteAngle = WriteAngle;
   meta_engfuncs.pfnWriteCoord = WriteCoord;
   meta_engfuncs.pfnWriteString = WriteString;
   meta_engfuncs.pfnWriteEntity = WriteEntity;

   memcpy (pengfuncsFromEngine, &meta_engfuncs, sizeof (enginefuncs_t));
   return TRUE;
}

// END of Metamod stuff




#ifdef _WIN32
BOOL WINAPI DllMain (HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
   return TRUE; // required Win32 DLL entry point
}
#endif


void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals)
{
   memcpy (&g_engfuncs, pengfuncsFromEngine, sizeof (enginefuncs_t));
   gpGlobals = pGlobals;

   // set the arrays to their default values
   memset (acknowledged_players, 0, sizeof (acknowledged_players));
   memset (acknowledgment_time, 0, sizeof (acknowledgment_time));
}


void ClientDisconnect (edict_t *pEntity)
{
   int entity_index = ENTINDEX (pEntity); // get the client's index

   acknowledged_players[entity_index] = FALSE; // reset this player's acknowledgment
   acknowledgment_time[entity_index] = 0; // don't send any message to a disconnected client

   RETURN_META (MRES_IGNORED);
}


void ClientPutInServer (edict_t *pEntity)
{
   int entity_index = ENTINDEX (pEntity); // get the client's index

   // if this client is a bot, make it acknowledge automatically
   if (pEntity->v.flags & FL_FAKECLIENT)
   {
      acknowledged_players[entity_index] = TRUE; // set this bot's acknowledgment
      acknowledgment_time[entity_index] = 0; // don't display message to bots
   }
   else
   {
      acknowledged_players[entity_index] = FALSE; // reset this player's acknowledgment
      acknowledgment_time[entity_index] = gpGlobals->time + 3.0; // display message time
   }

   RETURN_META (MRES_IGNORED);
}


void ClientCommand (edict_t *pEntity)
{
   int entity_index = ENTINDEX (pEntity); // get the sender's index

   // if this player hasn't acknowledged the game rules yet...
   if (!acknowledged_players[entity_index])
   {
      // if it's a bot, make it acknowledge automatically
      if (pEntity->v.flags & FL_FAKECLIENT)
      {
         acknowledged_players[entity_index] = TRUE; // set this bot's acknowledgment
         acknowledgment_time[entity_index] = 0; // reset this bot's message time
         RETURN_META (MRES_IGNORED); // bots issue whatever command they want
      }

      // else if he's acknowledging right now, allow him to pass commands
      if (strcmp (CMD_ARGV (0), CVAR_GET_STRING ("acknowledgment_string")) == 0)
      {
         acknowledged_players[entity_index] = TRUE; // set this player's acknowledgment
         acknowledgment_time[entity_index] = 0; // reset this player's message time

         // is this player forced to choose a team ?
         if (CVAR_GET_FLOAT ("player_force_team") > 0)
         {
            engClientCommand (pEntity, "chooseteam\n"); // make him display the team selection menu
            engClientCommand (pEntity, "menuselect %s\n", CVAR_GET_STRING ("player_force_team")); // make him join the right team
         }
      }

      RETURN_META (MRES_SUPERCEDE); // ignore this command
   }

   // else if it's a choose team command issued by a real client...
   else if ((strcmp (CMD_ARGV (0), "chooseteam") == 0) && (pEntity->v.flags & FL_CLIENT) && !(pEntity->v.flags & FL_FAKECLIENT))
      if (CVAR_GET_FLOAT ("player_force_team") > 0)
         RETURN_META (MRES_SUPERCEDE); // if this player was forced to choose his team, ignore this command

   RETURN_META (MRES_IGNORED); // other players issue whatever command they want
}


void StartFrame (void)
{
   int playerindex;

   // if the round has been restarted...
   if (round_end)
   {
      round_end = FALSE; // reset end of round signal

      // if the map tactics message has to be sent...
      if (CVAR_GET_FLOAT ("send_map_message") > 0)
         SendMapMessage (); // send acknowledged players the map message
   }

   // cycle through all players
   for (playerindex = 1; playerindex <= gpGlobals->maxClients; playerindex++)
   {
      edict_t *pPlayer = INDEXENT (playerindex);

      if (!pPlayer || pPlayer->free)
         continue; // skip invalid players

      // is it time to send this player the welcome message ?
      if (!acknowledged_players[playerindex] && (acknowledgment_time[playerindex] > 0) && (acknowledgment_time[playerindex] < gpGlobals->time))
      {
         SendWelcomeMessage (pPlayer); // send him the welcome message
         acknowledgment_time[playerindex] = gpGlobals->time + 3.0; // update in 3 seconds
      }
   }

   RETURN_META (MRES_IGNORED);
}


void engClientCommand (edict_t *pEdict, char *szFmt, ...)
{
   va_list ap;
   char buf[1024];

   va_start (ap, szFmt);
   vsnprintf (buf, sizeof (buf), szFmt, ap);
   va_end (ap);

   (*g_engfuncs.pfnClientCommand) (pEdict, buf);

   RETURN_META (MRES_IGNORED);
}


void MessageBegin (int msg_dest, int msg_type, const float *pOrigin, edict_t *ed)
{
   // is this message actually directed to somebody ?
   if (!FNullEnt (ed))
   {
      // is this message for a real client AND this client hasn't acknowledged the game rules yet
      // AND this message is for displaying some kind of menu ?
      if ((ed->v.flags & FL_CLIENT) && !(ed->v.flags & FL_FAKECLIENT) && (!acknowledged_players[ENTINDEX (ed)])
          && ((GET_USER_MSG_ID (PLID, "VGUIMenu", NULL) == msg_type) || (GET_USER_MSG_ID (PLID, "ShowMenu", NULL) == msg_type)))
      {
         ignore_message = TRUE; // then discard it
         RETURN_META (MRES_SUPERCEDE);
      }

      // is it a round end message ?
      if (GET_USER_MSG_ID (PLID, "RoundTime", NULL))
         round_end = TRUE; // catch new round message
   }

   RETURN_META (MRES_IGNORED);
}


void MessageEnd (void)
{
   // has this message to be ignored ?
   if (ignore_message)
   {
      ignore_message = FALSE; // the message is over now, so reset the ignore_message flag
      RETURN_META (MRES_SUPERCEDE); // don't send this packet
   }

   RETURN_META (MRES_IGNORED); // send the packet
}


void WriteByte (int iValue)
{
   // has this message to be ignored ?
   if (ignore_message)
      RETURN_META (MRES_SUPERCEDE); // don't send this packet

   RETURN_META (MRES_IGNORED); // send the packet
}


void WriteChar (int iValue)
{
   // has this message to be ignored ?
   if (ignore_message)
      RETURN_META (MRES_SUPERCEDE); // don't send this packet

   RETURN_META (MRES_IGNORED); // send the packet
}


void WriteShort (int iValue)
{
   // has this message to be ignored ?
   if (ignore_message)
      RETURN_META (MRES_SUPERCEDE); // don't send this packet

   RETURN_META (MRES_IGNORED); // send the packet
}


void WriteLong (int iValue)
{
   // has this message to be ignored ?
   if (ignore_message)
      RETURN_META (MRES_SUPERCEDE); // don't send this packet

   RETURN_META (MRES_IGNORED); // send the packet
}


void WriteAngle (float flValue)
{
   // has this message to be ignored ?
   if (ignore_message)
      RETURN_META (MRES_SUPERCEDE); // don't send this packet

   RETURN_META (MRES_IGNORED); // send the packet
}


void WriteCoord (float flValue)
{
   // has this message to be ignored ?
   if (ignore_message)
      RETURN_META (MRES_SUPERCEDE); // don't send this packet

   RETURN_META (MRES_IGNORED); // send the packet
}


void WriteString (const char *sz)
{
   // has this message to be ignored ?
   if (ignore_message)
      RETURN_META (MRES_SUPERCEDE); // don't send this packet

   RETURN_META (MRES_IGNORED); // send the packet
}


void WriteEntity (int iValue)
{
   // has this message to be ignored ?
   if (ignore_message)
      RETURN_META (MRES_SUPERCEDE); // don't send this packet

   RETURN_META (MRES_IGNORED); // send the packet
}


void SendWelcomeMessage (edict_t *pClient)
{
   FILE *fp;

   if (pClient->v.flags & FL_FAKECLIENT)
      return; // don't send anything to bots

   // gets welcome message, print a default one if file void or not found
   if ((fp = fopen ("cstrike/welcome.msg", "r")) != NULL)
   {
      fread (welcome_text, 512, 1, fp);
      fclose (fp);
   }

   // send the welcome message to this client
   (*g_engfuncs.pfnMessageBegin) (MSG_ONE, SVC_TEMPENTITY, NULL, pClient);
   (*g_engfuncs.pfnWriteByte) (TE_TEXTMESSAGE);
   (*g_engfuncs.pfnWriteByte) (1); // channel
   (*g_engfuncs.pfnWriteShort) (-8192); // x coordinates * 8192
   (*g_engfuncs.pfnWriteShort) (-8192); // y coordinates * 8192
   (*g_engfuncs.pfnWriteByte) (0); // effect (fade in/out)
   (*g_engfuncs.pfnWriteByte) (255); // initial RED
   (*g_engfuncs.pfnWriteByte) (255); // initial GREEN
   (*g_engfuncs.pfnWriteByte) (255); // initial BLUE
   (*g_engfuncs.pfnWriteByte) (1); // initial ALPHA
   (*g_engfuncs.pfnWriteByte) (255); // effect RED
   (*g_engfuncs.pfnWriteByte) (255); // effect GREEN
   (*g_engfuncs.pfnWriteByte) (255); // effect BLUE
   (*g_engfuncs.pfnWriteByte) (1); // effect ALPHA
   (*g_engfuncs.pfnWriteShort) (256); // fade-in time in seconds * 256
   (*g_engfuncs.pfnWriteShort) (512); // fade-out time in seconds * 256
   (*g_engfuncs.pfnWriteShort) (512); // hold time in seconds * 256
   (*g_engfuncs.pfnWriteString) (welcome_text); // message string
   (*g_engfuncs.pfnMessageEnd) (); // end
}


void SendMapMessage (void)
{
   FILE *fp;
   char filename[64];

   // build the file path according to the map name
   sprintf (filename, "cstrike/maps/%s.msg", STRING (gpGlobals->mapname));

   // gets map message
   if ((fp = fopen (filename, "r")) != NULL)
   {
      fread (map_text, 512, 1, fp);
      fclose (fp);
   }

   // cycle through all players
   for (int playerindex = 1; playerindex <= gpGlobals->maxClients; playerindex++)
   {
      edict_t *pClient = INDEXENT (playerindex);

      if (!pClient || pClient->free)
         continue; // skip invalid players

      if (pClient->v.flags & FL_FAKECLIENT)
         continue; // don't send anything to bots

      if (!acknowledged_players[playerindex])
         continue; // don't send anything to non-acknowledged players

      // send the map message to this client
      (*g_engfuncs.pfnMessageBegin) (MSG_ONE, SVC_TEMPENTITY, NULL, pClient);
      (*g_engfuncs.pfnWriteByte) (TE_TEXTMESSAGE);
      (*g_engfuncs.pfnWriteByte) (1); // channel
      (*g_engfuncs.pfnWriteShort) (-8192); // x coordinates * 8192
      (*g_engfuncs.pfnWriteShort) (-8192); // y coordinates * 8192
      (*g_engfuncs.pfnWriteByte) (0); // effect (fade in/out)
      (*g_engfuncs.pfnWriteByte) (255); // initial RED
      (*g_engfuncs.pfnWriteByte) (255); // initial GREEN
      (*g_engfuncs.pfnWriteByte) (255); // initial BLUE
      (*g_engfuncs.pfnWriteByte) (1); // initial ALPHA
      (*g_engfuncs.pfnWriteByte) (255); // effect RED
      (*g_engfuncs.pfnWriteByte) (255); // effect GREEN
      (*g_engfuncs.pfnWriteByte) (255); // effect BLUE
      (*g_engfuncs.pfnWriteByte) (1); // effect ALPHA
      (*g_engfuncs.pfnWriteShort) (256); // fade-in time in seconds * 256
      (*g_engfuncs.pfnWriteShort) (1024); // fade-out time in seconds * 256
      (*g_engfuncs.pfnWriteShort) (CVAR_GET_FLOAT ("mp_freezetime") * 256); // hold time in seconds * 256
      (*g_engfuncs.pfnWriteString) (map_text); // message string
      (*g_engfuncs.pfnMessageEnd) (); // end
   }
}


void GetGameDir (char *game_dir)
{
   // This function fixes the erratic behaviour caused by the use of the GET_GAME_DIR engine
   // macro, which returns either an absolute directory path, or a relative one, depending on
   // whether the game server is run standalone or not. This one always return a RELATIVE path.

   unsigned char length, fieldstart, fieldstop;

   GET_GAME_DIR (game_dir); // call the engine macro and let it mallocate for the char pointer

   length = strlen (game_dir); // get the length of the returned string
   length--; // ignore the trailing string terminator

   // format the returned string to get the last directory name
   fieldstop = length;
   while (((game_dir[fieldstop] == '\\') || (game_dir[fieldstop] == '/')) && (fieldstop > 0))
      fieldstop--; // shift back any trailing separator

   fieldstart = fieldstop;
   while ((game_dir[fieldstart] != '\\') && (game_dir[fieldstart] != '/') && (fieldstart > 0))
      fieldstart--; // shift back to the start of the last subdirectory name

   if ((game_dir[fieldstart] == '\\') || (game_dir[fieldstart] == '/'))
      fieldstart++; // if we reached a separator, step over it

   // now copy the formatted string back onto itself character per character
   for (length = fieldstart; length <= fieldstop; length++)
      game_dir[length - fieldstart] = game_dir[length];
   game_dir[length - fieldstart] = 0; // terminate the string

   return;
}
